/*
	Vibrator Control Protocol Daemon
	Copyright (C) 2003  Matthias Bruestle <m@mbsks.franken.de>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <arpa/inet.h>	/* inet_ntoa					*/
#include <netdb.h>		/* host to IP resolution		*/
#include <netinet/in.h>	/* Internet address structures	*/
#include <signal.h>		/* define signal(), etc.		*/
#include <stdio.h>		/* Basic I/O routines			*/
#include <stdlib.h>		/* Basic routines				*/
#include <string.h>		/* String handling				*/
#include <syslog.h>		/* Syslog functions				*/
#include <sys/socket.h>	/* socket interface functions	*/
#include <sys/time.h>	/* for timeout values			*/
#include <sys/types.h>	/* standard system types		*/
#include <sys/wait.h>	/* define wait(), etc.			*/
#include <time.h>		/* time, ctime					*/
#include <unistd.h>		/* for table size calculations	*/

#include <sio.h>		/* Serial I/O					*/
#include <vcp.h>		/* VCP defines					*/

int running;

/* signal handler */
void catch_child(int sig_num)
{
	/* when we get here, we know there's a zombie child waiting */
	int child_status;

	wait( &child_status );

	running = 0;
}

/* and the child process forking code... */
int fork_worker( S_VCP_STATE *vs, char *d )
{
	int child_pid;
	int ret;

	if( running ) {
		syslog( LOG_WARNING, "Worker called, but already one running\n" );
		close( vs->cs );
		return FALSE;
	}

	child_pid = fork();
	switch( child_pid ) {
	case -1:	/* fork() failed */
		syslog( LOG_WARNING, "Fork failed.\n" );
		return FALSE;
	case 0:		/* child */
		break;
	default:	/* inside parent process */
		running = 1;
		return TRUE;
	}

	if( !vcp_init( vs, d ) ) {
		write( vs->cs, VCP_RETCODE_ERROR, strlen(VCP_RETCODE_ERROR) );
		exit( 0 );
	}

	/* call protocol handler */
	ret = vcp_control( vs );

	SIO_Close( vs->si );

	/* terminate */
	syslog( LOG_INFO, "Disconnect.\n" );
	close( vs->cs );
	exit( ret );
}

int main( int argc, char *argv[] )
{
	int		i;					/* index counter for loop operations */
	int		rc;					/* system calls return value storage */
	int		s;					/* socket descriptor */
	struct sockaddr_in	sa;		/* Internet address struct */
	struct sockaddr_in	csa;	/* client's address struct */
	int		size_csa;			/* size of client's address struct */
	int		dsize;				/* size of file descriptors table */
	FILE	*f;
	char	*c;
	struct hostent		*hp;
	long	timevar;

	S_VCP_STATE vs;

	running = 0;
	memset( &vs, 0, sizeof(vs) );
	vs.intv=NULL;
	vs.currintv=NULL;

	if( argc!=3 ) {
		fprintf( stderr, "vcpd: Wrong number of arguments.\n" );
		exit( 1 );
	}

	if( strcmp( argv[1], "ser" )==0 ) {
		fprintf( stderr, "vcpd: Vibrator type serial.\n" );
		vs.type=VIB_TYPE_SER;
	} else if( strcmp( argv[1], "dummy" )==0 ) {
		fprintf( stderr, "vcpd: Vibrator type dummy.\n" );
		vs.type=VIB_TYPE_DUMMY;
	} else {
		fprintf( stderr, "vcpd: Wrong vibrator type.\n" );
		exit( 1 );
	}

	if( gethostname( vs.hostname, sizeof(vs.hostname)-1 )!=0 ) {
		strncpy( vs.hostname, "unknown", sizeof(vs.hostname)-1 );
	}

	if( (f=fopen( VCP_CONFIG, "r" ))==NULL ) {
		fprintf( stderr, "vcpd: Error opening config file.\n" );
		exit( 1 );
	}

	/* Read user name */
	if( fgets( &vs.username[0], sizeof(vs.username), f )==NULL ) {
		fprintf( stderr, "vcpd: Error reading user name.\n" );
		exit( 1 );
	}
	c=strchr( &vs.username[0], '\n' );
	if( c!=NULL ) *c=0;

	/* Read auth data */
	if( fgets( &vs.buffer[0], sizeof(vs.buffer), f )==NULL ) {
		fprintf( stderr, "vcpd: Error reading authentication data.\n" );
		exit( 1 );
	}
	/* Copy login */
	vs.auth.login[sizeof(vs.auth.login)-1]=0;
	strncpy( &vs.auth.login[0], &vs.buffer[0], sizeof(vs.auth.login)-1 );
	c=strchr( &vs.auth.login[0], ' ' );
	if( c!=NULL ) *c=0;
	/* Copy password */
	c=strchr( &vs.buffer[0], ' ' );
	while( *c==' ' ) c++;
	vs.auth.password[sizeof(vs.auth.password)-1]=0;
	strncpy( &vs.auth.password[0], c, sizeof(vs.auth.password)-1 );
	c=strchr( &vs.auth.password[0], ' ' );
	if( c!=NULL ) *c=0;
	c=strchr( &vs.auth.password[0], '\n' );
	if( c!=NULL ) *c=0;
	/* Check size */
	if( (strlen(&vs.auth.login[0])==0) || (strlen(&vs.auth.password[0])==0) ) {
		fprintf( stderr, "vcpd: Error reading authentication data.\n" );
		exit( 1 );
	}

	openlog( "vcpd", LOG_CONS|LOG_PID, LOG_DAEMON );
	syslog( LOG_INFO, "Daemon started. (Protocol version %s)", VCP_VERSION );

	/* Fork and kill the parent */
	if( fork() != 0 )
		return 0;

	/* calculate size of file descriptors table */
	dsize = getdtablesize();

	/* close all file descriptors, except our communication socket		*/
	/* this is done to avoid blocking on tty operations and such.		*/
	for (i=0; i<dsize; i++) close(i);

	if( chdir( "/" ) ) return 1;
	if( setsid()==-1 ) return 1;

	/* Do it a second time to avoid zombie processes */
	if( fork() != 0 )
		return 0;

	/* define the signal handler for the CHLD signal */
	signal( SIGCHLD, catch_child );

	/* initiate machine's Internet address structure */
	/* first clear out the struct, to avoid garbage	 */
	memset( &sa, 0, sizeof(sa) );
	/* Using Internet address family */
	sa.sin_family = AF_INET;
	/* copy port number in network byte order */
	sa.sin_port = htons(VCP_PORT);
	/* we will accept cnnections coming through any IP	*/
	/* address that belongs to our host, using the		*/
	/* INADDR_ANY wild-card.							*/
	sa.sin_addr.s_addr = INADDR_ANY;

	/* allocate a free socket				  */
	/* Internet address family, Stream socket */
	s = socket( AF_INET, SOCK_STREAM, 0 );
	if (s < 0) {
		syslog( LOG_ERR, "socket: allocation failed." );
		return 1;
	}

	/* bind the socket to the newly formed address */
	rc = bind( s, (struct sockaddr *)&sa, sizeof(sa) );

	/* check there was no error */
	if( rc ) {
		syslog( LOG_ERR, "bind error." );
		return 1;
	}

	/* ask the system to listen for incoming connections		*/
	/* to the address we just bound. specify that up to			*/
	/* 5 pending connection requests will be queued by the		*/
	/* system, if we are not directly awaiting them using		*/
	/* the accept() system call, when they arrive.				*/
	rc = listen( s, 5 );

	/* check there was no error */
	if( rc ) {
		syslog( LOG_ERR, "listen error." );
	}

	/* remember size for later usage */
	size_csa = sizeof(csa);

	/* enter an accept-write-close infinite loop */
	for(;;) {
		/* the accept() system call will wait for a		*/
		/* connection, and when one is established, a	*/
		/* new socket will be created to form it, and	*/
		/* the csa variable will hold the address		*/
		/* of the Client that just connected to us.		*/
		/* the old socket, s, will still be available	*/
		/* for future accept() statements.				*/
		vs.cs = accept(s, (struct sockaddr *)&csa, &size_csa);

		/* check for errors. if any, enter accept mode again */
		if( vs.cs < 0 )
			continue;

		time (&timevar);

		/* Look up the host information for the remote host
		 * that we have connected with.	 Its internet address
		 * was returned by the accept call, in the main
		 * daemon loop above.
		 */
		hp = gethostbyaddr ((char *) &csa.sin_addr, sizeof (struct in_addr),
			csa.sin_family);

		if (hp == NULL) {
			/* The information is unavailable for the remote
			 * host.  Just format its internet address to be
			 * printed out in the logging information.	The
			 * address will be shown in "internet dot format".
			 */
			c = inet_ntoa(csa.sin_addr);
		} else {
			c = hp->h_name;	/* point to host's name */
		}

		if( c==NULL ) {
			syslog( LOG_NOTICE, "Connect from unknown at %s.", ctime(&timevar) );
		} else {
			syslog( LOG_NOTICE, "Connect from %s at %s.", c, ctime(&timevar) );
		}

		if( running ) {
			syslog( LOG_WARNING, "Already connected." );
			/* Already one connection */
			snprintf( vs.buffer, VCP_BUFLEN, VCP_RETCODE_BUSY );
			write( vs.cs, vs.buffer, strlen(vs.buffer) );
		} else {
			/* ok, we got a new connection. do the job... */
			fork_worker( &vs, argv[2] );
		}

		/* now close the connection */
		close( vs.cs );
	}

	return 0;
}

